/* ReadTheDocs theme colors */
/* code adapted from scanpy custom css and from https://github.com/godotengine/godot-docs/blob/master/_static/css/custom.css */

/* navbar color */
.wy-side-nav-search, .wy-nav-top, .wy-nav-side {
    background: #003262;
}
.wy-nav-content {
    max-width: 1100px;
}
.wy-side-nav-search input[type="text"] { border-width: 0 }

/* Heading color in navbar */
.wy-menu-vertical header,
.wy-menu-vertical p.caption {
    color: #C4820E;

    /* Improves the appearance of uppercase text */
    letter-spacing: 0.75px;
}


/* Custom classes */

.small { font-size:40% }
.smaller, .pr { font-size:70% }


/* Custom classes with bootstrap buttons */

/* Formatting of RTD markup: rubrics and sidebars and admonitions */

/* rubric */
.rst-content p.rubric {
    margin-bottom: 6px;
    font-weight: normal;
}
.rst-content p.rubric::after { content: ":" }

/* sidebar */
.rst-content .sidebar {
    /* margin: 0px 0px 0px 12px; */
    padding-bottom: 0px;
}
.rst-content .sidebar p {
    margin-bottom: 12px;
}
.rst-content .sidebar p,
.rst-content .sidebar ul,
.rst-content .sidebar dl {
    font-size: 13px;
}

/* less space after bullet lists in admonitions like warnings and notes */
.rst-content .section .admonition ul {
    margin-bottom: 6px;
}


/* override table width restrictions */
@media screen and (min-width: 767px) {

    .wy-table-responsive table td {
       /* !important prevents the common CSS stylesheets from overriding
          this as on RTD they are loaded after this stylesheet */
       white-space: normal !important;
    }

    .wy-table-responsive {
       overflow: visible !important;
    }
 }


/* Code: literals and links */

.rst-content tt.literal,
.rst-content code.literal {
    color: #404040;
}
/* slim font weight for non-link code */
.rst-content tt:not(.xref),
.rst-content code:not(.xref),
.rst-content *:not(a) > tt.xref,
.rst-content *:not(a) > code.xref,
.rst-content dl:not(.docutils) code
{
    font-weight: normal;
}
.rst-content a > tt.xref,
.rst-content a > code.xref,
.rst-content dl:not(.docutils) a > tt.xref,
.rst-content dl:not(.docutils) a > code.xref
{
    font-weight: bold;  /* underline looks clumsy, in particular with buttons and
                           other hyperlinks, which don't come with underlines */
}


/* Just one box for annotation code for a less noisy look */

.rst-content .annotation {
    padding: 2px 5px;
    background-color: white;
    border: 1px solid #e1e4e5;
}
.rst-content .annotation tt,
.rst-content .annotation code {
    padding: 0 0;
    background-color: transparent;
    border: 0 solid transparent;
}


/* Parameter lists */

/* Mimick rubric style used for other headings */
/* TODO: once scanpydoc adds classes, also change return types like this */
.rst-content dl:not(.docutils) dl > dt {
    font-weight: bold;
    background: #edf0f2;
    border-left: solid 3px #003262;
    margin: 0 0 12px;
    padding: 8px 5px 8px 8px;
}
/* Parameters contain <strong> parts and don’t need bold font */
.rst-content dl.field-list dl > dt { font-weight: unset }
/* Add colon between return tuple element name and type */
.rst-content dl:not(.docutils) dl > dt .classifier::before { content: ' : ' }

/* Function headers */

.rst-content dl:not(.docutils) dt {
    background: #edf0f2;
    color: #404040;
    border-top: solid 3px #003262;
    padding: 8px 5px 8px 8px;
}
